/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *     Alexander Diewald (fortiss GmbH) - backport to AF3 & Integration
 *                                        improvements
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import javafx.beans.property.SimpleObjectProperty;
import javafx.event.EventHandler;
import javafx.scene.Node;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.SelectionMode;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeTableCell;
import javafx.scene.control.TreeTableColumn;
import javafx.scene.control.TreeTableColumn.CellEditEvent;
import javafx.scene.control.TreeTableView;
import javafx.scene.control.cell.TextFieldTreeTableCell;

/**
 * A JavaFX {@link TreeTableView} based on an {@link DynamicTreeContentProviderBase} and
 * {@link DynamicTreeTableUIProviderBase}. This class is intended to be used as a wrapper for
 * {@link TreeTableView} and its rather complex expert API.
 * <P>
 * This class currently supports only a static tree-table layout, i.e., there is no way to alter the
 * columns unless the developer changes the underlying {@link TreeTableView} directly. Note that if
 * a column is removed the index provided to the label provider might not be correct anymore.
 * <P>
 * Furthermore, the selection model is currently fixed to be row selection.
 */
public final class DynamicTreeTableViewer<T> extends DynamicTreeViewerBase<T> {
	/** The {@link TreeTableView} control to be managed. */
	private final TreeTableView<T> view;
	/** The UI provider of this tree-table. */
	private final DynamicTreeTableUIProviderBase uiProvider;

	/** Constructor. */
	public DynamicTreeTableViewer(TreeTableView<T> view, T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeTableUIProviderBase uiProvider) {
		super(contentProvider);
		this.uiProvider = uiProvider;
		// construct view
		this.view = view;
		DynamicTreeItem<T> rootItem = new DynamicTreeItem<T>(root, this);
		view.setRoot(rootItem);
		view.setShowRoot(showRoot);
		view.setEditable(true);
		view.getSelectionModel().setSelectionMode(SelectionMode.MULTIPLE);
		rootItem.update();
		// expand to reveal (+1 if root node is not shown
		expandItem(rootItem, showRoot ? revealLevel : revealLevel + 1);
	}

	/** Constructor. */
	public DynamicTreeTableViewer(T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeTableUIProviderBase uiProvider) {
		this(new TreeTableView<T>(), root, showRoot, revealLevel, contentProvider, uiProvider);
	}

	/** Updates the viewer content. */
	public void update() {
		// wild cast works: see constructor
		DynamicTreeItem<T> rootItem = (DynamicTreeItem<T>)view.getRoot();
		rootItem.update();
	}

	/** Expands items up to the given level. */
	private void expandItem(TreeItem<T> parentItem, int revealLevel) {
		if(revealLevel <= 0) {
			return;
		}
		parentItem.setExpanded(true);
		revealLevel--;
		for(TreeItem<T> child : parentItem.getChildren()) {
			expandItem(child, revealLevel);
		}
	}

	/**
	 * Creates non-editable column displaying the labels as defined by the
	 * {@link DynamicTreeTableUIProviderBase} used as a basis for specific column types.
	 */
	private TreeTableColumn<T, String> createColumn(String headerLabel, int prefWidth) {
		int num = view.getColumns().size();
		TreeTableColumn<T, String> column = new TreeTableColumn<>(headerLabel);

		column.setPrefWidth(prefWidth);
		column.setCellValueFactory(param -> {
			Object data = param.getValue().getValue();
			return new SimpleObjectProperty<String>(uiProvider.getLabel(data, num));
		});

		view.getColumns().add(column);

		return column;
	}

	/**
	 * Adds a non-editable column to the table part of the view. The labels, context menus and icons
	 * are shown as defined in the {@link DynamicTreeTableUIProviderBase}.
	 */
	// TODO(AD): Please add a flag to make the column read-only or read-writable.
	public void addColumn(String headerLabel, int prefWidth) {
		TreeTableColumn<T, String> column = createColumn(headerLabel, prefWidth);

		column.setCellFactory(param -> {
			TreeTableCell<T, String> cell = new TreeTableCell<T, String>() {
				@Override
				protected void updateItem(String item, boolean empty) {
					super.updateItem(item, empty);
					ContextMenu menu = null;
					Node icon = null;
					int index = view.getColumns().size() - 1;
					if(!empty && item != null) {
						T data = this.getTreeTableRow().getItem();
						menu = uiProvider.createContextMenu(data, index);
						icon = uiProvider.getIconNode(data, index);
					}
					this.setContextMenu(menu);
					this.setGraphic(icon);
				}
			};
			cell.textProperty().bind(cell.itemProperty());
			return cell;
		});
	}

	/**
	 * Adds an editable column to the table part of the view. The labels defined in the
	 * {@link DynamicTreeTableUIProviderBase} are used as predefined values. The
	 * {@link EventHandler} is executed after a cell has been successfully edited.
	 */
	// TODO(AD): This method is redundant and oversimplified (due to the limitation to a text field
	// editor). It would be better to change the method "addColumn(...)" to return the created
	// column. The client code would be pretty much the same with that change.
	public void addColumn(String headerLabel, int prefWidth,
			EventHandler<CellEditEvent<T, String>> eventHandler) {
		TreeTableColumn<T, String> column = createColumn(headerLabel, prefWidth);

		column.setCellFactory(TextFieldTreeTableCell.<T>forTreeTableColumn());
		column.setOnEditCommit(eventHandler);
	}

	/** Returns the underlying {@link TreeTableView}. */
	public TreeTableView<T> getControl() {
		return view;
	}
}
